"""Loads configuration files

This module loads json configuration files from disk,
and combines them with the default settings,
to create one dict that contains all parameters.
It also checks that all parameters exists, and that
no new parameters have been added by accident.
"""

from os.path import dirname, join
import logging
import json
import jsonschema

logger = logging.getLogger(__name__)

if int(jsonschema.__version__[0]) < 3:  # pragma: no cover
    logger.warning(
        "Jsonschema %s found, but at least 3.0.0 is required to check configuration. Skipping the check.",
        jsonschema.__version__,
    )
    hasJsonSchema = False
else:
    hasJsonSchema = True


def get_configuration_for_instrument(instrument, plot=None):
    local = dirname(__file__)
    instrument = str(instrument)
    if instrument in ["pyreduce", None]:
        fname = join(local, "settings", f"settings_pyreduce.json")
    else:
        fname = join(local, "settings", f"settings_{instrument.upper()}.json")

    config = load_config(fname, instrument)

    if plot is not None:
        for key, value in config.items():
            if isinstance(config[key], dict) and "plot" in config[key].keys():
                config[key]["plot"] = plot

    return config


def load_config(configuration, instrument, j=0):
    if configuration is None:
        logger.info(
            "No configuration specified, using default values for this instrument"
        )
        config = get_configuration_for_instrument(instrument, plot=False)
    elif isinstance(configuration, dict):
        if instrument in configuration.keys():
            config = configuration[str(instrument)]
        elif (
            "__instrument__" in configuration.keys()
            and configuration["__instrument__"] == str(instrument).upper()
        ):
            config = configuration
        else:
            raise KeyError("This configuration is for a different instrument")
    elif isinstance(configuration, list):
        config = configuration[j]
    elif isinstance(configuration, str):
        config = configuration

    if isinstance(config, str):
        logger.info("Loading configuration from %s", config)
        try:
            with open(config) as f:
                config = json.load(f)
        except FileNotFoundError:
            fname = dirname(__file__)
            fname = join(fname, "settings", config)
            with open(fname) as f:
                config = json.load(f)

    # Combine instrument specific settings, with default values
    settings = read_config()
    settings = update(settings, config)

    # If it doesn't raise an Exception everything is as expected
    validate_config(settings)
    logger.debug("Configuration succesfully validated")

    return settings


def update(dict1, dict2, check=True):
    """
    Update entries in dict1 with entries of dict2 recursively,
    i.e. if the dict contains a dict value, values inside the dict will
    also be updated

    Parameters
    ----------
    dict1 : dict
        dict that will be updated
    dict2 : dict
        dict that contains the values to update
    check : bool
        If True, will check that the keys from dict2 exist in dict1 already.
        Except for those contained in field "instrument"

    Returns
    -------
    dict1 : dict
        the updated dict

    Raises
    ------
    KeyError
        If dict2 contains a key that is not in dict1
    """
    # Instrument is a 'special' section as it may include any number of values
    # In that case we don't want to raise an error for new keys
    exclude = ["instrument"]
    for key, value in dict2.items():
        if check and key not in dict1.keys():
            raise KeyError(f"{key} is not contained in dict1")
        if isinstance(value, dict):
            dict1[key] = update(dict1[key], value, check=key not in exclude)
        else:
            dict1[key] = value
    return dict1


def read_config(fname="settings_pyreduce.json"):
    """Read the configuration file from disk

    If no filename is given it will load the default configuration.
    The configuration file must be a json file.

    Parameters
    ----------
    fname : str, optional
        Filename of the configuration. By default "settings_pyreduce.json",
        i.e. the default configuration

    Returns
    -------
    config : dict
        The read configuration file
    """
    this_dir = dirname(__file__)
    fname = join(this_dir, "settings", fname)

    with open(fname) as file:
        settings = json.load(file)
        return settings


def validate_config(config):
    """Test that the input configuration complies with the expected schema

    Since it requires features from jsonschema 3+, it will only run if that is installed.
    Otherwise show a warning but continue. This is incase some other module needs an earlier,
    jsonschema (looking at you jwst).

    If the function runs through without raising an exception, the check was succesful or skipped.

    Parameters
    ----------
    config : dict
        Configurations to check

    Raises
    ------
    ValueError
        If there is a problem with the configuration.
        Usually that means a setting has an unallowed value.
    """
    if not hasJsonSchema:  # pragma: no cover
        # Can't check with old version
        return
    fname = "settings_schema.json"
    this_dir = dirname(__file__)
    fname = join(this_dir, "settings", fname)

    with open(fname) as f:
        schema = json.load(f)
    try:
        jsonschema.validate(schema=schema, instance=config)
    except jsonschema.ValidationError as ve:
        logger.error("Configuration failed validation check.\n%s", ve.message)
        raise ValueError(ve.message)



import pytest
from os.path import dirname, join

import numpy as np
from astropy.io import fits

from pyreduce import instruments
from pyreduce.combine_frames import combine_bias


def test_bias(instrument, mode, files, mask):
    if len(files["bias"]) == 0:
        pytest.skip(f"No bias files for instrument {instrument}")

    bias, bhead = combine_bias(
        files["bias"], instrument, mode, window=50, mask=mask
    )

    assert isinstance(bias, np.ma.masked_array)
    assert isinstance(bhead, fits.Header)

    assert bias.ndim == 2
    assert bias.shape[0] == mask.shape[0]
    assert bias.shape[1] == mask.shape[1]


def test_only_one_file(instrument, mode, files, mask):
    if len(files["bias"]) == 0:
        pytest.skip(f"No bias files for instrument {instrument}")

    files = [files["bias"][0]]
    bias, bhead = combine_bias(files, instrument, mode, window=50, mask=mask)

    assert isinstance(bias, np.ma.masked_array)
    assert isinstance(bhead, fits.Header)

    assert bias.ndim == 2
    assert bias.shape[0] == mask.shape[0]
    assert bias.shape[1] == mask.shape[1]


def test_no_data_files():
    with pytest.raises(FileNotFoundError):
        combine_bias([], "", "")


def test_wrong_data_type():
    with pytest.raises(TypeError):
        combine_bias(None, None, "")

    with pytest.raises(ValueError):
        combine_bias([None], None, "")


def test_simple_input(tempfiles):
    n = 2
    files = tempfiles[:n]

    for i in range(n):
        data = np.full((100, 100), i, dtype=float)
        fits.writeto(files[i], data)

    bias, bhead = combine_bias(files, "common", "", extension=0)

    assert isinstance(bias, np.ndarray)
    assert bias.shape[0] == 100
    assert bias.shape[1] == 100
    assert np.all(bias == sum(range(n)) / n)



import pytest
import numpy as np

from pyreduce.estimate_background_scatter import estimate_background_scatter


def test_scatter(flat, orders, settings):
    # The background scatter step in reduce possibly uses a
    # different set of files for the image
    # However it should still be able to create a scatter fit
    # from the flat image as is done here
    img, _ = flat
    orders, column_range = orders
    settings = settings["scatter"]
    settings["sigma_cutoff"] = settings["scatter_cutoff"]
    del settings["scatter_cutoff"]
    del settings["bias_scaling"]

    if img is None:
        pytest.skip("Need flat")

    scatter = estimate_background_scatter(
        img, orders, column_range=column_range, **settings
    )

    degree = settings["scatter_degree"]
    if np.isscalar(degree):
        degree = [degree, degree]

    assert isinstance(scatter, np.ndarray)
    assert scatter.ndim == 2
    assert scatter.shape[0] == degree[0] + 1
    assert scatter.shape[1] == degree[1] + 1


def test_simple():
    img = np.full((100, 100), 10.0)
    orders = np.array([[0, 25], [0, 50], [0, 75]])

    scatter = estimate_background_scatter(img, orders, scatter_degree=0, plot=False)

    assert isinstance(scatter, np.ndarray)
    assert scatter.ndim == 2
    assert scatter.shape[0] == 1
    assert scatter.shape[1] == 1

    assert np.allclose(scatter[0, 0], 10.0)


def test_scatter_degree():
    img = np.full((100, 100), 10.0)
    orders = np.full((2, 2), 1.0)

    estimate_background_scatter(img, orders, scatter_degree=0)

    with pytest.raises(ValueError):
        estimate_background_scatter(img, orders, scatter_degree=-1)

    estimate_background_scatter(img, orders, scatter_degree=(2, 2))

    with pytest.raises(AssertionError):
        estimate_background_scatter(img, orders, scatter_degree=(1,))

    with pytest.raises(AssertionError):
        estimate_background_scatter(img, orders, scatter_degree=(3, 2, 1))

    with pytest.raises(ValueError):
        estimate_background_scatter(img, orders, scatter_degree=(2, -1))



"""
Handles instrument specific info for the HARPS spectrograph

Mostly reading data from the header
"""
import os.path
import glob
import logging
from datetime import datetime
import fnmatch
import re

import numpy as np
from astropy.io import fits
from astropy.time import Time
from dateutil import parser

from .common import getter, InstrumentWithModes, observation_date_to_night
from .filters import Filter

logger = logging.getLogger(__name__)


class MCDONALD(InstrumentWithModes):
    def _convert_time_deg(self, v):
        v = [float(s) for s in v.split(":")]
        v = v[0] + v[1] / 60 + v[2] / 3600
        return v

    def add_header_info(self, header, mode, **kwargs):
        """ read data from header and add it as REDUCE keyword back to the header """
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works

        header = super().add_header_info(header, mode, **kwargs)
        info = self.load_info()
        get = getter(header, info, mode)

        header["e_orient"] = get("orientation", 0)

        trimsec = get("trimsec")

        if trimsec is not None:
            pattern = r"\[(\d*?):(\d*?),(\d*?):(\d*?)\]"
            res = re.match(pattern, trimsec)
            prescan_x = int(res[1]) + 1
            overscan_x = int(res[2])
            prescan_y = int(res[3])
            overscan_y = int(res[4])
        else:
            prescan_x = 2
            overscan_x = 2048
            prescan_y = 2
            overscan_y = 2047

        header["e_xlo"] = prescan_x
        header["e_xhi"] = overscan_x

        header["e_ylo"] = prescan_y
        header["e_yhi"] = overscan_y

        amp = get("amplifier")
        gain = info["gain"].format(amplifier=amp)
        readnoise = info["readnoise"].format(amplifier=amp)

        header["e_gain"] = get(gain, 1)
        header["e_readn"] = get(readnoise, 0)
        header["e_exptim"] = get("exposure_time", 0)

        header["e_sky"] = get("sky", 0)
        header["e_drk"] = get("dark", 0)
        header["e_backg"] = header["e_gain"] * (header["e_drk"] + header["e_sky"])

        header["e_imtype"] = get("observation_type")
        header["e_ctg"] = get("observation_type")

        obs_date = get("date")
        ut = get("universal_time")
        dark_time = get("dark_time")
        ra = get("ra")
        dec = get("dec")

        if ra is not None:
            ra = self._convert_time_deg(ra)
        if dec is not None:
            dec = self._convert_time_deg(dec)
        if ut is not None and dark_time is not None:
            tmid = self._convert_time_deg(ut) + dark_time / 2
        else:
            tmid = 0
        if obs_date is not None:
            jd = Time(obs_date).mjd + tmid + 0.5
        else:
            jd = 0

        header["e_ra"] = ra
        header["e_dec"] = dec
        header["e_jd"] = jd

        header["e_obslon"] = self._convert_time_deg(info["longitude"])
        header["e_obslat"] = self._convert_time_deg(info["latitude"])
        header["e_obsalt"] = info["altitude"]

        return header

    def get_wavecal_filename(self, header, mode, **kwargs):
        """ Get the filename of the wavelength calibration config file """
        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_2D.npz".format(instrument="harps", mode=mode)
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname



"""
Handles instrument specific info for the HARPS spectrograph

Mostly reading data from the header
"""
import os.path
import glob
import logging
from datetime import datetime
import fnmatch
import json

import numpy as np
from astropy.io import fits
from dateutil import parser

from .common import getter, Instrument, observation_date_to_night
from .filters import Filter

logger = logging.getLogger(__name__)


class CRIRES_PLUS(Instrument):
    def __init__(self):
        super().__init__()
        self.filters["lamp"] = Filter(self.info["id_lamp"])
        self.filters["band"] = Filter(self.info["id_band"])
        self.filters["decker"] = Filter(self.info["id_decker"])

    def add_header_info(self, header, mode, **kwargs):
        """ read data from header and add it as REDUCE keyword back to the header """
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)
        info = self.load_info()

        return header

    def get_expected_values(self, target, night, mode):
        expectations = super().get_expected_values(target, night)

        band, decker = mode.rsplit("_", 1)
        for key in expectations.keys():
            if key == "bias":
                continue
            expectations[key]["band"] = band
            expectations[key]["decker"] = decker

        expectations["wavecal"]["lamp"] = self.info["id_lamp_wavecal"]
        expectations["freq_comb"]["lamp"] = self.info["id_lamp_etalon"]

        return expectations

    # def sort_files(self, input_dir, target, night, mode):
    #     """
    #     Sort a set of fits files into different categories
    #     types are: bias, flat, wavecal, orderdef, spec

    #     Parameters
    #     ----------
    #     input_dir : str
    #         input directory containing the files to sort
    #     target : str
    #         name of the target as in the fits headers
    #     night : str
    #         observation night, possibly with wildcards
    #     mode : str
    #         instrument mode
    #     Returns
    #     -------
    #     files_per_night : list[dict{str:dict{str:list[str]}}]
    #         a list of file sets, one entry per night, where each night consists of a dictionary with one entry per setting,
    #         each fileset has five lists of filenames: "bias", "flat", "order", "wave", "spec", organised in another dict
    #     nights_out : list[datetime]
    #         a list of observation times, same order as files_per_night
    #     """

    #     info = self.load_info()
    #     target = target.upper().replace("-", "")
    #     instrument = "CRIRES"

    #     # Try matching with nights
    #     try:
    #         night = parser.parse(night).date()
    #         individual_nights = [night]
    #     except ValueError:
    #         # if the input night can't be parsed, use all nights
    #         # Usually the case if wildcards are involved
    #         individual_nights = "all"

    #     # find all fits files in the input dir(s)
    #     input_dir = input_dir.format(
    #         instrument=instrument.upper(), target=target, mode=mode, night=night
    #     )
    #     files = glob.glob(input_dir + "/*.fits")
    #     files += glob.glob(input_dir + "/*.fits.gz")
    #     files = np.array(files)

    #     # Initialize arrays
    #     # observed object
    #     ob = np.zeros(len(files), dtype="U20")
    #     # observation type, bias, flat, spec, etc.
    #     ty = np.zeros(len(files), dtype="U20")
    #     # instrument mode, e.g. red, blue
    #     mo = np.zeros(len(files), dtype="U20")
    #     band = np.zeros(len(files), dtype="U20")
    #     decker = np.zeros(len(files), dtype="U20")
    #     lamp = np.zeros(len(files), dtype="U20")

    #     se = np.zeros(len(files), dtype="U20")

    #     # observed night, parsed into a datetime object
    #     ni = np.zeros(len(files), dtype=datetime)
    #     # instrument, used for observation
    #     it = np.zeros(len(files), dtype="U20")

    #     for i, f in enumerate(files):
    #         h = fits.open(f)[0].header
    #         ob[i] = h.get(info["target"], "")
    #         ty[i] = h.get(info["observation_type"], "")
    #         mo[i] = h.get(info["id_mode"])
    #         band[i] = h.get(info["id_band"], "").replace(r"/", "_")
    #         decker[i] = h.get(info["id_decker"], "").upper()
    #         lamp_tmp = h.get(info["id_lamp"], "")
    #         if len(lamp_tmp) != 0:
    #             lamp[i] = lamp_tmp[0]
    #         se[i] = band[i] + "_" + decker[i]
    #         ni_tmp = h.get(info["date"], "")
    #         it[i] = h.get(info["instrument"], "")
    #         # Sanitize input
    #         ni[i] = observation_date_to_night(ni_tmp)
    #         ob[i] = ob[i].replace("-", "")

    #     if isinstance(individual_nights, str) and individual_nights == "all":
    #         individual_nights = np.unique(ni)
    #         logger.info(
    #             "Can't parse night %s, use all %i individual nights instead",
    #             night,
    #             len(individual_nights),
    #         )

    #     files_per_night = []
    #     nights_out = []
    #     for ind_night in individual_nights:
    #         # Select files for this night, this instrument, this instrument mode
    #         selection = (ni == ind_night) & (it == instrument)

    #         # Find all unique setting keys for this night and target
    #         # Only look at the settings of observation files
    #         # match_ty = np.array([fnmatch.fnmatch(t, info["id_spec"]) for t in ty])
    #         # match_ob = np.array([fnmatch.fnmatch(t, target) for t in ob])

    #         if mode != "":
    #             keys = np.unique(se[selection & (se == mode)])
    #         else:
    #             keys = np.unique(se[selection])

    #         files_this_night = {}
    #         for key in keys:
    #             select = selection & (se == key)

    #             # find all relevant files for this setting
    #             # bias ignores the setting
    #             files_this_night = {
    #                 "bias": files[(ty == info["id_bias"]) & selection],
    #                 "flat": files[(ty == info["id_flat"]) & select],
    #                 "orders": files[(ty == info["id_flat"]) & select],
    #                 "wavecal": files[
    #                     (ty == info["id_wave"])
    #                     & (lamp == info["id_lamp_wavecal"])
    #                     & select
    #                 ],
    #                 "freq_comb": files[
    #                     (ty == info["id_wave"])
    #                     & (lamp == info["id_lamp_etalon"])
    #                     & select
    #                 ],
    #                 "science": [],
    #             }
    #             files_this_night["curvature"] = (
    #                 files_this_night["freq_comb"]
    #                 if len(files_this_night["freq_comb"]) != 0
    #                 else files_this_night["wavecal"]
    #             )
    #             files_this_night["scatter"] = files_this_night["orders"]

    #             files_per_night.append(
    #                 (
    #                     {"night": ind_night, "key": key, "target": target},
    #                     files_this_night,
    #                 )
    #             )

    #     return files_per_night

    def get_wavecal_filename(self, header, mode, **kwargs):
        """ Get the filename of the wavelength calibration config file """
        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_2D.npz".format(instrument="harps", mode=mode)
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname



import pytest
import numpy as np
from pyreduce.cwrappers import slitfunc, slitfunc_curved


def test_slitfunc_vert():
    img = np.full((100, 100), 1, dtype=float)
    ycen = np.full(100, 0, dtype=float)
    lambda_sp = 0
    lambda_sf = 0.1
    osample = 1

    # Run it once the way it is supposed to
    slitfunc(img, ycen, lambda_sp, lambda_sf, osample)

    # Then try different incompatible inputs, which have to be caught before going to the C code
    with pytest.raises(AssertionError):
        slitfunc(None, ycen, lambda_sp, lambda_sf, osample)
    with pytest.raises(ValueError):
        slitfunc("bla", ycen, lambda_sp, lambda_sf, osample)

    with pytest.raises(AssertionError):
        slitfunc(img, None, lambda_sp, lambda_sf, osample)
    with pytest.raises(ValueError):
        slitfunc(img, "blub", lambda_sp, lambda_sf, osample)

    with pytest.raises(TypeError):
        slitfunc(img, ycen, None, lambda_sf, osample)
    with pytest.raises(ValueError):
        slitfunc(img, ycen, "bla", lambda_sf, osample)
    with pytest.raises(TypeError):
        slitfunc(img, ycen, lambda_sp, None, osample)
    with pytest.raises(ValueError):
        slitfunc(img, ycen, lambda_sp, "bla", osample)
    with pytest.raises(TypeError):
        slitfunc(img, ycen, lambda_sp, lambda_sf, None)
    with pytest.raises(ValueError):
        slitfunc(img, ycen, lambda_sp, lambda_sf, "bla")

    # Then try different sizes for img and ycen
    with pytest.raises(AssertionError):
        ycen_bad = np.full(50, 1, dtype=float)
        slitfunc(img, ycen_bad, lambda_sp, lambda_sf, osample)

    with pytest.raises(AssertionError):
        slitfunc(img, ycen, lambda_sp, lambda_sf, osample=0)
    with pytest.raises(AssertionError):
        slitfunc(img, ycen, lambda_sp, -1, osample)
    with pytest.raises(AssertionError):
        slitfunc(img, ycen, -1, lambda_sf, osample)


def test_slitfunc_curved():
    img = np.full((100, 100), 1)
    ycen = np.full(100, 50)
    tilt = np.full(100, 0)
    shear = np.full(100, 0)
    lambda_sp = 0
    lambda_sf = 0.1
    osample = 1
    yrange = (49, 50)

    # Run it once the way it is supposed to
    slitfunc_curved(img, ycen, tilt, shear, lambda_sp, lambda_sf, osample, yrange)
    slitfunc_curved(img, ycen, 1, 0.01, lambda_sp, lambda_sf, osample, yrange)

    # Then try different incompatible inputs, which have to be caught before going to the C code
    with pytest.raises(AssertionError):
        slitfunc_curved(None, ycen, tilt, shear, lambda_sp, lambda_sf, osample, yrange)
    with pytest.raises(ValueError):
        slitfunc_curved("bla", ycen, tilt, shear, lambda_sp, lambda_sf, osample, yrange)

    with pytest.raises(AssertionError):
        slitfunc_curved(img, None, tilt, shear, lambda_sp, lambda_sf, osample, yrange)
    with pytest.raises(ValueError):
        slitfunc_curved(img, "blub", tilt, shear, lambda_sp, lambda_sf, osample, yrange)

    with pytest.raises(AssertionError):
        slitfunc_curved(img, ycen, tilt, None, lambda_sp, lambda_sf, osample, yrange)
    with pytest.raises(ValueError):
        slitfunc_curved(img, ycen, tilt, "boo", lambda_sp, lambda_sf, osample, yrange)

    with pytest.raises(TypeError):
        slitfunc_curved(img, ycen, tilt, shear, None, lambda_sf, osample, yrange)
    with pytest.raises(ValueError):
        slitfunc_curved(img, ycen, tilt, shear, "bla", lambda_sf, osample, yrange)
    with pytest.raises(TypeError):
        slitfunc_curved(img, ycen, tilt, shear, lambda_sp, None, osample, yrange)
    with pytest.raises(ValueError):
        slitfunc_curved(img, ycen, tilt, shear, lambda_sp, "bla", osample, yrange)
    with pytest.raises(TypeError):
        slitfunc_curved(img, ycen, tilt, shear, lambda_sp, lambda_sf, None, yrange)
    with pytest.raises(ValueError):
        slitfunc_curved(img, ycen, tilt, shear, lambda_sp, lambda_sf, "bla", yrange)

    # Then try different sizes for img and ycen
    with pytest.raises(AssertionError):
        ycen_bad = np.full(50, 0, dtype=float)
        slitfunc_curved(img, ycen_bad, tilt, shear, lambda_sp, lambda_sf, osample, yrange)

    with pytest.raises(AssertionError):
        tilt_bad = np.full(50, 0, dtype=float)
        slitfunc_curved(img, ycen, tilt_bad, shear, lambda_sp, lambda_sf, osample, yrange)

    with pytest.raises(AssertionError):
        shear_bad = np.full(50, 0, dtype=float)
        slitfunc_curved(img, ycen, tilt, shear_bad, lambda_sp, lambda_sf, osample, yrange)

    with pytest.raises(AssertionError):
        slitfunc_curved(img, ycen, tilt, shear, lambda_sp, lambda_sf, 0, yrange)
    with pytest.raises(AssertionError):
        slitfunc_curved(img, ycen, tilt, shear, lambda_sp, -1, osample, yrange)
    with pytest.raises(AssertionError):
        slitfunc_curved(img, ycen, tilt, shear, -1, lambda_sf, osample, yrange)




import numpy as np
from fnmatch import fnmatch
import re
from datetime import datetime, date
from dateutil import parser
import logging
from astropy.time import Time
from astropy import units as u

logger = logging.getLogger(__name__)


class Filter:
    def __init__(
        self,
        keyword,
        dtype="U20",
        wildcards=False,
        regex=False,
        flags=0,
        unique=True,
        ignorecase=True,
    ):
        self.keyword = keyword
        self.dtype = dtype
        self.wildcards = wildcards
        self.regex = regex
        self.flags = flags
        self.data = []
        self.unique = unique
        self.ignorecase = ignorecase

        if self.ignorecase and not self.flags & re.IGNORECASE:
            self.flags += re.IGNORECASE

    def collect(self, header):
        if self.keyword is None:
            value = ""
        else:
            value = header.get(self.keyword)
        if value.__class__ == header.__class__:
            if len(value) > 0:
                value = value[0]
            else:
                value = ""
        self.data.append(value)
        return value

    def match(self, value):
        if self.keyword is None:
            result = np.full(len(self.data), False)
        else:
            try:
                if self.regex:
                    regex = re.compile(f"^(?:{value})$", flags=self.flags)
                elif self.wildcards:
                    regex = re.compile(fnmatch.translate(value), flags=self.flags)
                else:
                    regex = re.compile(value, flags=self.flags)

                result = [
                    regex.match(f) is not None if f is not None else False
                    for f in self.data
                ]
            except TypeError as ex:
                result = [f == value for f in self.data]
        result = np.asarray(result, dtype=bool)
        return result

    def classify(self, value):
        if self.unique:
            if value is not None and value != "":
                match = self.match(value)
                data = np.asarray(self.data)
                data = np.unique(data[match])
            else:
                data = set(self.data)
            data = [(d, self.match(d)) for d in data]
        else:
            if value is not None and value != "":
                match = self.match(value)
            else:
                match = np.full(len(self.data), True)
            data = [(value, match)]
        return data

    def clear(self):
        self.data = []


class InstrumentFilter(Filter):
    def __init__(self, keyword="INSTRUME", **kwargs):
        kwargs["dtype"] = "U20"
        kwargs["unique"] = False
        super().__init__(keyword, **kwargs)


class ObjectFilter(Filter):
    def __init__(self, keyword="OBJECT", **kwargs):
        kwargs["dtype"] = "U20"
        kwargs["unique"] = False
        super().__init__(keyword, **kwargs)


class NightFilter(Filter):
    def __init__(self, keyword="DATE-OBS", timeformat="fits", **kwargs):
        super().__init__(keyword, dtype=datetime, **kwargs)
        self.timeformat = timeformat

    @staticmethod
    def observation_date_to_night(observation_date):
        """Convert an observation timestamp into the date of the observation night
        Nights start at 12am and end at 12 am the next day
        """
        if observation_date.to_datetime().hour < 12:
            observation_date -= 1 * u.day
        return observation_date.to_datetime().date()

    def collect(self, header):
        value = header.get(self.keyword)
        if value is not None:
            value = Time(value, format=self.timeformat)
            value = NightFilter.observation_date_to_night(value)
        else:
            logger.warning(
                "Could not determine the observation date of %s, skipping it", header
            )
        self.data.append(value)
        return value

    def match(self, value):
        try:
            value = parser.parse(value).date()
        except Exception:
            pass
        match = super().match(value)
        return match


class ModeFilter(Filter):
    def __init__(
        self,
        keyword,
        dtype="U20",
        wildcards=False,
        regex=False,
        flags=0,
        unique=True,
        ignorecase=True,
        replacement={},
    ):
        super().__init__(
            keyword,
            dtype=dtype,
            wildcards=wildcards,
            regex=regex,
            flags=flags,
            unique=unique,
            ignorecase=ignorecase,
        )
        self.replacement = replacement

    def classify(self, value):
        data = super().classify(value)
        data = [
            (self.replacement[d] if d in self.replacement.keys() else d, m)
            for d, m in data
        ]
        return data



import pytest
import os
import sys

from pyreduce import reduce


def test_main(instrument, target, night, mode, input_dir, output_dir):
    output = reduce.main(
        instrument,
        target,
        night,
        mode,
        base_dir="",
        input_dir=input_dir,
        output_dir=output_dir,
        steps=(),
    )

    assert isinstance(output, list)
    assert len(output) >= 1
    assert "config" in output[0].keys()
    assert "files" in output[0].keys()

    # Test default options
    # Just just not find anything
    output = reduce.main(instrument, target, night, steps=())


@pytest.mark.skip
def test_run_all(instrument, target, night, mode, input_dir, output_dir, order_range):
    output = reduce.main(
        instrument,
        target,
        night,
        mode,
        base_dir="",
        input_dir=input_dir,
        output_dir=output_dir,
        order_range=order_range,
        steps="all",
    )


@pytest.mark.skip
def test_load_all(instrument, target, night, mode, input_dir, output_dir, order_range):
    output = reduce.main(
        instrument,
        target,
        night,
        mode,
        base_dir="",
        input_dir=input_dir,
        output_dir=output_dir,
        order_range=order_range,
        steps=["finalize"],
    )


def test_step_abstract(step_args):
    step = reduce.Step(*step_args, **{"plot": False})

    assert isinstance(step.dependsOn, list)
    assert isinstance(step.loadDependsOn, list)
    assert isinstance(step.prefix, str)
    assert isinstance(step.output_dir, str)

    with pytest.raises(NotImplementedError):
        step.load()

    with pytest.raises(NotImplementedError):
        step.run([])

    with pytest.raises(NotImplementedError):
        step.save()



"""
Handles instrument specific info for the HARPS spectrograph

Mostly reading data from the header
"""
import os.path
import glob
import logging
from datetime import datetime
import fnmatch
import json

import numpy as np
from astropy.io import fits
from dateutil import parser

from .common import getter, Instrument, observation_date_to_night

logger = logging.getLogger(__name__)


class JWST_MIRI(Instrument):
    def add_header_info(self, header, mode, **kwargs):
        """ read data from header and add it as REDUCE keyword back to the header """
        # "Normal" stuff is handled by the general version, specific changes to values happen here
        # alternatively you can implement all of it here, whatever works
        header = super().add_header_info(header, mode)
        info = self.load_info()
        return header

    def get_wavecal_filename(self, header, mode, **kwargs):
        """ Get the filename of the wavelength calibration config file """
        cwd = os.path.dirname(__file__)
        fname = "{instrument}_{mode}_2D.npz".format(instrument="harps", mode=mode)
        fname = os.path.join(cwd, "..", "wavecal", fname)
        return fname
